@@ -1,6 +1,6 @@
 ;;; Guile bytecode disassembler
 
-;;; Copyright (C) 2001, 2009, 2010, 2012, 2013, 2014, 2015 Free Software Foundation, Inc.
+;;; Copyright (C) 2001, 2009-2010, 2012-2015, 2017-2018 Free Software Foundation, Inc.
 ;;;
 ;;; This library is free software; you can redistribute it and/or
 ;;; modify it under the terms of the GNU Lesser General Public
@@ -24,6 +24,7 @@
   #:use-module (system vm debug)
   #:use-module (system vm program)
   #:use-module (system vm loader)
+  #:use-module (system base types internal)
   #:use-module (system foreign)
   #:use-module (rnrs bytevectors)
   #:use-module (ice-9 format)
@@ -71,6 +72,11 @@
       s
       (- s (ash 1 24))))
 
+(define (unpack-s12 s)
+  (if (zero? (logand s (ash 1 11)))
+      s
+      (- s (ash 1 12))))
+
 (define (unpack-s32 s)
   (if (zero? (logand s (ash 1 31)))
       s
@@ -96,6 +102,9 @@
             X8_F12_F12)
            #'((logand (ash word -8) #xfff)
               (ash word -20)))
+          ((X8_S12_Z12)
+           #'((logand (ash word -8) #xfff)
+              (unpack-s12 (ash word -20))))
           ((X8_S8_S8_S8
             X8_S8_S8_C8
             X8_S8_C8_S8)
@@ -112,9 +121,12 @@
            #'(word))
           ((N32 R32 L32 LO32)
            #'((unpack-s32 word)))
-          ((C8_C24)
+          ((C8_C24 C8_S24)
            #'((logand word #xff)
               (ash word -8)))
+          ((C16_C16)
+           #'((logand word #xffff)
+              (ash word -16)))
           ((B1_C7_L24)
            #'((not (zero? (logand word #x1)))
               (logand (ash word -1) #x7f)
@@ -177,6 +189,19 @@
 address of that offset."
   (+ (debug-context-base context) (* offset 4)))
 
+(define immediate-tag-annotations '())
+(define-syntax-rule (define-immediate-tag-annotation name pred mask tag)
+  (set! immediate-tag-annotations
+        (cons `((,mask ,tag) ,(symbol->string 'pred)) immediate-tag-annotations)))
+(visit-immediate-tags define-immediate-tag-annotation)
+
+(define heap-tag-annotations '())
+(define-syntax-rule (define-heap-tag-annotation name pred mask tag)
+  (set! heap-tag-annotations
+        (cons `((,mask ,tag) ,(symbol->string 'pred)) heap-tag-annotations)))
+
+(visit-heap-tags define-heap-tag-annotation)
+
 (define (code-annotation code len offset start labels context push-addr!)
   ;; FIXME: Print names for register loads and stores that correspond to
   ;; access to named locals.
@@ -190,31 +215,12 @@ address of that offset."
        (dereference-pointer (make-pointer addr)))))
 
   (match code
-    (((or 'br
-          'br-if-nargs-ne 'br-if-nargs-lt 'br-if-nargs-gt
-          'br-if-true 'br-if-null 'br-if-nil 'br-if-pair 'br-if-struct
-          'br-if-char 'br-if-eq 'br-if-eqv
-          'br-if-= 'br-if-< 'br-if-<= 'br-if-> 'br-if->=
-          'br-if-u64-= 'br-if-u64-< 'br-if-u64-<=
-          'br-if-u64-<-scm 'br-if-u64-<=-scm 'br-if-u64-=-scm
-          'br-if-u64->-scm 'br-if-u64->=-scm
-          'br-if-f64-= 'br-if-f64-< 'br-if-f64-<=
-          'br-if-f64-> 'br-if-f64->=
-          'br-if-logtest) _ ... target)
+    (((or 'j 'je 'jl 'jge 'jne 'jnl 'jnge) target)
      (list "-> ~A" (vector-ref labels (- (+ offset target) start))))
-    (('br-if-tc7 slot invert? tc7 target)
-     (list "~A -> ~A"
-           (let ((tag (case tc7
-                        ((5) "symbol?")
-                        ((7) "variable?")
-                        ((13) "vector?")
-                        ((15) "string?")
-                        ((53) "keyword?")
-                        ((77) "bytevector?")
-                        ((95) "bitvector?")
-                        (else (number->string tc7)))))
-             (if invert? (string-append "not " tag) tag))
-           (vector-ref labels (- (+ offset target) start))))
+    (('immediate-tag=? _ mask tag)
+     (assoc-ref immediate-tag-annotations (list mask tag)))
+    (('heap-tag=? _ mask tag)
+     (assoc-ref heap-tag-annotations (list mask tag)))
     (('prompt tag escape-only? proc-slot handler)
      ;; The H is for handler.
      (list "H -> ~A" (vector-ref labels (- (+ offset handler) start))))
@@ -229,14 +235,8 @@ address of that offset."
      (list "~a slot~:p" nlocals))
     (('reset-frame nlocals)
      (list "~a slot~:p" nlocals))
-    (('return-values nlocals)
-     (if (zero? nlocals)
-         (list "all values")
-         (list "~a value~:p" (1- nlocals))))
     (('bind-rest dst)
      (list "~a slot~:p" (1+ dst)))
-    (('tail-call nargs proc)
-     (list "~a arg~:p" nargs))
     (('make-closure dst target nfree)
      (let* ((addr (u32-offset->addr (+ offset target) context))
             (pdi (find-program-debug-info addr context))
@@ -244,6 +244,13 @@ address of that offset."
                       "anonymous procedure")))
        (push-addr! addr name)
        (list "~A at #x~X (~A free var~:p)" name addr nfree)))
+    (('load-label dst src)
+     (let* ((addr (u32-offset->addr (+ offset src) context))
+            (pdi (find-program-debug-info addr context))
+            (name (or (and pdi (program-debug-info-name pdi))
+                      "anonymous procedure")))
+       (push-addr! addr name)
+       (list "~A at #x~X" name addr)))
     (('call-label closure nlocals target)
      (let* ((addr (u32-offset->addr (+ offset target) context))
             (pdi (find-program-debug-info addr context))
@@ -251,7 +258,7 @@ address of that offset."
                       "anonymous procedure")))
        (push-addr! addr name)
        (list "~A at #x~X" name addr)))
-    (('tail-call-label nlocals target)
+    (('tail-call-label target)
      (let* ((addr (u32-offset->addr (+ offset target) context))
             (pdi (find-program-debug-info addr context))
             (name (or (and pdi (program-debug-info-name pdi))
@@ -263,22 +270,14 @@ address of that offset."
        (when (program? val)
          (push-addr! (program-code val) val))
        (list "~@Y" val)))
+    (((or 'throw/value 'throw/value+data) dst target)
+     (list "~@Y" (reference-scm target)))
     (('builtin-ref dst idx)
      (list "~A" (builtin-index->name idx)))
     (((or 'static-ref 'static-set!) _ target)
      (list "~@Y" (dereference-scm target)))
-    (((or 'free-ref 'free-set!) _ _ index)
-     (list "free var ~a" index))
     (('resolve-module dst name public)
      (list "~a" (if (zero? public) "private" "public")))
-    (('toplevel-box _ var-offset mod-offset sym-offset bound?)
-     (list "`~A'~A" (dereference-scm sym-offset)
-           (if bound? "" " (maybe unbound)")))
-    (('module-box _ var-offset mod-name-offset sym-offset bound?)
-     (let ((mod-name (reference-scm mod-name-offset)))
-       (list "`(~A ~A ~A)'~A" (if (car mod-name) '@ '@@) (cdr mod-name)
-             (dereference-scm sym-offset)
-             (if bound? "" " (maybe unbound)"))))
     (('load-typed-array dst type shape target len)
      (let ((addr (u32-offset->addr (+ offset target) context)))
        (list "~a bytes from #x~X" len addr)))
@@ -297,14 +296,7 @@ address of that offset."
             (match elt
               ((inst arg ...)
                (case inst
-                 ((br
-                   br-if-nargs-ne br-if-nargs-lt br-if-nargs-gt
-                   br-if-true br-if-null br-if-nil br-if-pair br-if-struct
-                   br-if-char br-if-tc7 br-if-eq br-if-eqv
-                   br-if-= br-if-< br-if-<= br-if-> br-if->= br-if-logtest
-                   br-if-u64-= br-if-u64-< br-if-u64-<=
-                   br-if-u64-<-scm br-if-u64-<=-scm br-if-u64-=-scm
-                   br-if-u64->-scm br-if-u64->=-scm)
+                 ((j je jl jge jne jnl jnge)
                   (match arg
                     ((_ ... target)
                      (add-label! (+ offset target) "L"))))
@@ -412,26 +404,14 @@ address of that offset."
        `(make-closure ,dst
                       ,(u32-offset->addr (+ offset target) context)
                       ,nfree))
+      (('load-label dst src)
+       `(load-label ,dst ,(u32-offset->addr (+ offset src) context)))
       (('make-non-immediate dst target)
        `(make-non-immediate ,dst ,(reference-scm target)))
       (('builtin-ref dst idx)
        `(builtin-ref ,dst ,(builtin-index->name idx)))
       (((or 'static-ref 'static-set!) dst target)
        `(,(car code) ,dst ,(dereference-scm target)))
-      (('toplevel-box dst var-offset mod-offset sym-offset bound?)
-       `(toplevel-box ,dst
-                      ,(dereference-scm var-offset)
-                      ,(dereference-scm mod-offset)
-                      ,(dereference-scm sym-offset)
-                      ,bound?))
-      (('module-box dst var-offset mod-name-offset sym-offset bound?)
-       (let ((mod-name (reference-scm mod-name-offset)))
-         `(module-box ,dst
-                      ,(dereference-scm var-offset)
-                      ,(car mod-name)
-                      ,(cdr mod-name)
-                      ,(dereference-scm sym-offset)
-                      ,bound?)))
       (_ code)))
   (let lp ((offset start) (seed seed))
     (cond
@@ -521,11 +501,11 @@ address of that offset."
 (define (instruction-has-fallthrough? code pos)
   (define non-fallthrough-set
     (static-opcode-set halt
-                       tail-call tail-call-label tail-call/shuffle
+                       throw throw/value throw/value+data
+                       tail-call tail-call-label
                        return-values
                        subr-call foreign-call continuation-call
-                       tail-apply
-                       br))
+                       j))
   (let ((opcode (logand (bytevector-u32-native-ref code pos) #xff)))
     (not (bitvector-ref non-fallthrough-set opcode))))
 
@@ -534,8 +514,7 @@ address of that offset."
     (syntax-case x ()
       ((_ name opcode kind word0 word* ...)
        (let ((symname (syntax->datum #'name)))
-         (if (or (memq symname '(br prompt))
-                 (string-prefix? "br-" (symbol->string symname)))
+         (if (memq symname '(prompt j je jl jge jne jnl jnge))
              (let ((offset (* 4 (length #'(word* ...)))))
                #`(vector-set!
                   jump-parsers
@@ -560,13 +539,13 @@ address of that offset."
     (define (stack-effect-parser name)
       (case name
         ((push)
-         #'(lambda (code pos size) (+ size 1)))
+         #'(lambda (code pos size) (and size (+ size 1))))
         ((pop)
-         #'(lambda (code pos size) (- size 1)))
+         #'(lambda (code pos size) (and size (- size 1))))
         ((drop)
          #'(lambda (code pos size)
              (let ((count (ash (bytevector-u32-native-ref code pos) -8)))
-               (- size count))))
+               (and size (- size count)))))
         ((alloc-frame reset-frame)
          #'(lambda (code pos size)
              (let ((nlocals (ash (bytevector-u32-native-ref code pos) -8)))
@@ -590,10 +569,14 @@ address of that offset."
                                   #xfff))
                    (nlocals (ash (bytevector-u32-native-ref code pos) -20)))
                (+ nargs nlocals))))
-        ((call call-label)
-         #'(lambda (code pos size) #f))
-        ((tail-call tail-call-label tail-call/shuffle tail-apply)
+        ((call call-label tail-call tail-call-label expand-apply-argument)
          #'(lambda (code pos size) #f))
+        ((shuffle-down)
+         #'(lambda (code pos size)
+             (let ((from (logand (ash (bytevector-u32-native-ref code pos) -8)
+                                 #xfff))
+                   (to (ash (bytevector-u32-native-ref code pos) -20)))
+               (and size (- size (- from to))))))
         (else
          #f)))
     (syntax-case x ()
@@ -623,10 +606,11 @@ address of that offset."
                                   (lambda ()
                                     (disassemble-one code (/ pos 4)))
                                 (lambda (len elt)
+                                  (define frame-size 3)
                                   (match elt
                                     ((_ proc . _)
-                                     (let lp ((slot (- proc 2)))
-                                       (if (< slot nslots-in)
+                                     (let lp ((slot (- proc frame-size)))
+                                       (if (and nslots-in (< slot nslots-in))
                                            (cons slot (lp (1+ slot)))
                                            '())))))))))
                  (vector-set! clobber-parsers opcode parse)))
@@ -644,7 +628,9 @@ address of that offset."
                                       ((X8_F24 X8_F12_F12)
                                        #'(list dst))
                                       (else
-                                       #'(list (- nslots-out 1 dst)))))))))))
+                                       #'(if nslots-out
+                                             (list (- nslots-out 1 dst))
+                                             '()))))))))))
               (vector-set! clobber-parsers opcode parse)))
          (else (error "unexpected instruction kind" #'kind)))))))
 
